﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.Diagnostics;

namespace Achiral.Extension
{
    public static partial class MakeWorld
    {
        public static void ForEach<T>(this IEnumerable<T> source, Action<T> action)
        {
            if (source == null) throw new ArgumentNullException("source");

            foreach (var item in source)
            {
                action(item);
            }
        }
        public static void ForEach<T>(this IEnumerable<T> source, Action<T, int> action)
        {
            if (source == null) throw new ArgumentNullException("source");

            checked
            {
                int index = 0;
                foreach (var item in source)
                {
                    action(item, index);
                    ++index;
                }
            }
        }
        public static string Join(this IEnumerable<string> source, string separator)
        {
            if (source == null) throw new ArgumentNullException("source");

            return string.Join(separator, source.ToArray());
        }
        public static IEnumerable<TResult> SelectWithFirst<T, TResult>(this IEnumerable<T> source, Func<T, T, TResult> func)
        {
            if (source == null) throw new ArgumentNullException("source");

            bool isFirst = false;
            var first = default(T);
            foreach (var item in source)
            {
                if (isFirst)
                {
                    first = item;
                    isFirst = false;
                }
                else
                {
                    yield return func(first, item);
                }
            }
        }

        public static IEnumerable<TResult> ZipWith<T1, T2, TResult>(this IEnumerable<T1> source1, IEnumerable<T2> source2, Func<T1, T2, TResult> func)
        {
            if (source1 == null) throw new ArgumentNullException("source1");
            if (source2 == null) throw new ArgumentNullException("source2");

            var enumerator1 = default(IEnumerator<T1>);
            var enumerator2 = default(IEnumerator<T2>);
            try
            {
                while (true)
                {
                    if (!enumerator1.MoveNext()) yield break;
                    if (!enumerator2.MoveNext()) yield break;

                    yield return func(enumerator1.Current, enumerator2.Current);
                }
            }
            finally
            {
                if (enumerator2 != null) enumerator2.Dispose();
                if (enumerator1 != null) enumerator1.Dispose();
            }
        }

        public static IEnumerable<TResult> ZipWith<T1, T2, TResult>(this IEnumerable<T1> source1, IEnumerable<T2> source2, Func<T1, T2, int, TResult> func)
        {
            if (source1 == null) throw new ArgumentNullException("source1");
            if (source2 == null) throw new ArgumentNullException("source2");

            var enumerator1 = default(IEnumerator<T1>);
            var enumerator2 = default(IEnumerator<T2>);
            try
            {
                checked
                {
                    int index = 0;
                    while (true)
                    {
                        if (!enumerator1.MoveNext()) yield break;
                        if (!enumerator2.MoveNext()) yield break;

                        yield return func(enumerator1.Current, enumerator2.Current, index);
                        ++index;
                    }
                }
            }
            finally
            {
                if (enumerator2 != null) enumerator2.Dispose();
                if (enumerator1 != null) enumerator1.Dispose();
            }
        }

        /// <see>http://d.hatena.ne.jp/siokoshou/20071227#c1198746342</see>
        public static IEnumerable<T> Cascade<T>(this IEnumerable<T> source, Func<T, IEnumerable<T>> func)
        {
            if (source == null) throw new ArgumentNullException("source");

            var enumeratorStack = new Stack<IEnumerator<T>>();
            try
            {
                var top = source.GetEnumerator();
                while (true)
                {
                    if (top.MoveNext())
                    {
                        yield return top.Current;
                        enumeratorStack.Push(top);
                        top = func(top.Current).GetEnumerator();
                    }
                    else
                    {
                        if (enumeratorStack.Count <= 0)
                        {
                            yield break;
                        }
                        top = enumeratorStack.Pop();
                    }
                }
            }
            finally
            {
                enumeratorStack.ForEach(e => e.Dispose());
            }
        }

        public static IEnumerable<TResult> Cascade<T, TResult>(this IEnumerable<T> source, Func<T, IEnumerable<T>> func, Func<T, int, TResult> selector)
        {
            if (source == null) throw new ArgumentNullException("source");

            var enumeratorStack = new Stack<IEnumerator<T>>();
            try
            {
                var top = source.GetEnumerator();
                while (true)
                {
                    if (top.MoveNext())
                    {
                        yield return selector(top.Current, enumeratorStack.Count);
                        enumeratorStack.Push(top);
                        top = func(top.Current).GetEnumerator();
                    }
                    else
                    {
                        if (enumeratorStack.Count <= 0)
                        {
                            yield break;
                        }
                        top = enumeratorStack.Pop();
                    }
                }

            }
            finally
            {
                enumeratorStack.ForEach(e => e.Dispose());
            }
        }

        public static IEnumerable<T> Trace<T>(this IEnumerable<T> source)
        {
            if (source == null) throw new ArgumentNullException("source");

            foreach (var item in source)
            {
                System.Diagnostics.Trace.WriteLine(item);
                yield return item;
            }
        }

        public static IEnumerable<T> Trace<T>(this IEnumerable<T> source, string category)
        {
            if (source == null) throw new ArgumentNullException("source");

            foreach (var item in source)
            {
                System.Diagnostics.Trace.WriteLine(item, category);
                yield return item;
            }
        }

        public static IEnumerable<T> Trace<T>(this IEnumerable<T> source, Func<T, string> selector)
        {
            if (source == null) throw new ArgumentNullException("source");

            foreach (var item in source)
            {
                System.Diagnostics.Trace.WriteLine(selector(item));
                yield return item;
            }
        }

        public static IEnumerable<T> Trace<T>(this IEnumerable<T> source, Func<T, string> selector, string category)
        {
            if (source == null) throw new ArgumentNullException("source");

            foreach (var item in source)
            {
                System.Diagnostics.Trace.WriteLine(selector(item), category);
                yield return item;
            }
        }

        public static IEnumerable<T> Trace<T>(this IEnumerable<T> source, Func<T, int, string> selector)
        {
            if (source == null) throw new ArgumentNullException("source");

            checked
            {
                int index = 0;
                foreach (var item in source)
                {
                    System.Diagnostics.Trace.WriteLine(selector(item, index));
                    yield return item;
                }
            }
        }

        public static IEnumerable<T> Trace<T>(this IEnumerable<T> source, Func<T, int, string> selector, string category)
        {
            if (source == null) throw new ArgumentNullException("source");

            checked
            {
                int index = 0;
                foreach (var item in source)
                {
                    System.Diagnostics.Trace.WriteLine(selector(item, index), category);
                    yield return item;
                    ++index;
                }
            }
        }

        public static void ConsoleWriteLine<T>(this IEnumerable<T> source)
        {
            if (source == null) throw new ArgumentNullException("source");

            foreach (var item in source)
            {
                Console.WriteLine(item);
            }
        }
        public static void ConsoleWriteLine<T>(this IEnumerable<T> source, Func<T, string> messageSelector)
        {
            if (source == null) throw new ArgumentNullException("source");

            foreach (var item in source)
            {
                Console.WriteLine(messageSelector(item));
            }
        }
        public static void ConsoleWriteLine<T>(this IEnumerable<T> source, string format, Func<T, object[]> argumentsSelector)
        {
            if (source == null) throw new ArgumentNullException("source");

            foreach (var item in source)
            {
                Console.WriteLine(format, argumentsSelector(item));
            }
        }
        public static void ConsoleWriteLine<T>(this IEnumerable<T> source, string format, params Func<T, object>[] selectors)
        {
            if (source == null) throw new ArgumentNullException("source");

            foreach (var item in source)
            {
                Console.WriteLine(format, selectors.Select(func => func(item)));
            }
        }

    }


}